// Interaktiver Brainfuck-Interpreter
// (c) Holger Guenther (hg-info@web.de), Doebeln, 10.05.2012

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License Version 2 as published
// by the Free Software Foundation. See license.txt for full details.

unit BFThread;

interface

uses
  Windows, Classes;

type
  TBFThread = class (TThread)
  private
    FSource  : string;
    FData    : string;
    FOutput  : string;
    FNewSrc  : boolean;
    FLock    : TRTLCriticalSection;
    procedure Lock;
    procedure Unlock;
    procedure SyncStart;
    procedure SyncResult;
  protected
    procedure Execute; override;
  public
    constructor Create;
    destructor Destroy; override;
    procedure SetSource (const ASource,AData: string);
  end;

implementation

uses
  MainFrm;

constructor TBFThread.Create;
begin
  inherited Create (false);
  FreeOnTerminate:=false;
  Priority:=tpLower;
  InitializeCriticalSection (FLock);
  FSource:='';
  FData:='';
  FOutput:='';
  FNewSrc:=false;
end;

destructor TBFThread.Destroy;
begin
  DeleteCriticalSection (FLock);
  inherited Destroy;
end;

procedure TBFThread.Lock;
begin
  EnterCriticalSection (FLock);
end;

procedure TBFThread.Unlock;
begin
  LeaveCriticalSection (FLock);
end;

procedure TBFThread.Execute;
var Cells    : array of cardinal;
    NumCells : cardinal;
    src,dat  : string;
    sp,dp,cp : cardinal;
    i,l      : cardinal;
begin
  NumCells:=256;
  SetLength (Cells,NumCells);
  while not Terminated do begin
    Sleep (1);
    if not FNewSrc then Continue;
    // Werte uebernehmen
    Lock;
    src:=FSource;
    dat:=FData;
    FNewSrc:=false;
    Unlock;
    // Brainfuck-Interpreter ausfuehren
    Synchronize (SyncStart);
    FillChar (Cells[0],NumCells*SizeOf (Cells[0]),0);
    FOutput:='';
    l:=Length (src);
    cp:=0;
    dp:=0;
    sp:=0;
    while (sp<l) and not (FNewSrc or Terminated) do begin
      Inc (sp);
      case src[sp] of
        '>': cp:=(cp+1) mod NumCells;
        '<': cp:=(cp+NumCells-1) mod NumCells;
        '+': if Cells[cp]<Cardinal (MaxLongint) then Inc (Cells[cp]);
        '-': if Cells[cp]>0 then Dec (Cells[cp]);
        '.': FOutput:=FOutput+Char (Byte (Cells[cp]));
        ',': if dp<Cardinal (Length (dat)) then begin
               Inc (dp);
               Cells[cp]:=Byte (dat[dp]);
               end
             else Cells[cp]:=0;
        '[': if Cells[cp]=0 then begin
               i:=1;
               while (sp<l) and (i>0) do begin
                 Inc (sp);
                 if src[sp]='[' then
                   Inc (i)
                 else if src[sp]=']' then
                   Dec (i);
               end;
             end;
        ']': if Cells[cp]<>0 then begin
               i:=1;
               while (sp>0) and (i>0) do begin
                 Dec (sp);
                 if sp<1 then break;
                 if src[sp]=']' then
                   Inc (i)
                 else if src[sp]='[' then
                   Dec (i);
               end;
             end;
      end;
    end;
    if not (FNewSrc or Terminated) then Synchronize (SyncResult);
  end;
  SetLength (Cells,0);
end;

procedure TBFThread.SetSource (const ASource,AData: string);
begin
  Lock;
  try
    if (ASource<>FSource) or (AData<>FData) then begin
      FSource:=ASource;
      FData:=AData;
      FNewSrc:=true;
    end;
  finally
    Unlock;
  end;
end;

procedure TBFThread.SyncStart;
begin
  if MainForm is TMainForm then MainForm.OnBFThreadStart (Self);
end;

procedure TBFThread.SyncResult;
begin
  if MainForm is TMainForm then MainForm.OnBFThreadResult (Self,FOutput);
end;

end.

